from typing import Tuple


class ClientAuth:
	def __init__(self, settings):
		self.settings = settings
	
	def get_seq_number(self) -> bool:
		"""
		Returns the initial sequence number.
		"""
		raise NotImplementedError()
	
	def signing_needed(self) -> bool:
		"""
		Checks if integrity protection was negotiated
		"""
		raise NotImplementedError()
	
	def encryption_needed(self) -> bool:
		"""
		Checks if confidentiality was negotiated
		"""
		raise NotImplementedError()
	
	def get_session_key(self) -> bytes:
		"""
		Returns the negotiated session key
		"""
		raise NotImplementedError()
				
	async def sign(self, data:bytes, seq_number:int) -> bytes:
		"""
		Signs a message. 
		"""
		raise NotImplementedError()
		
	async def encrypt(self, data:bytes, seq_number:int) -> bytes:
		"""
		Encrypts a message. 
		"""
		raise NotImplementedError()
		
	async def decrypt(self, data:bytes) -> bytes:
		"""
		Decrypts message. Also performs integrity checking.
		"""
		raise NotImplementedError()
	
	
	async def authenticate(self, token:bytes) -> Tuple[bytes, bool, Exception]:
		"""
		This function is called (multiple times depending on the protocol) to perform authentication.

		returns a tuple of (token, to_continue, error)
		"""
		return None, None, None


from asyauth.protocols.ntlm.structures.fields import Fields
from asyauth.protocols.ntlm.structures.negotiate_flags import NegotiateFlags
from asyauth.protocols.ntlm.structures.version import Version, WindowsMajorVersion, WindowsMinorVersion

# LDAP doesnt seem to support sign-only. either no seal nor sign nor always_sign OR include seal.
NTLMClientTemplates = {
		"Windows10_15063" : {
			'flags'            :  NegotiateFlags.NEGOTIATE_KEY_EXCH|
								  NegotiateFlags.NEGOTIATE_128|
								  NegotiateFlags.NEGOTIATE_VERSION|
								  NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|
								  NegotiateFlags.NEGOTIATE_NTLM|
								  NegotiateFlags.REQUEST_TARGET|
								  NegotiateFlags.NEGOTIATE_UNICODE,
			'version'          : Version.construct(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_10, minor = WindowsMinorVersion.WINDOWS_MINOR_VERSION_0, build = 15063 ),
			'domain_name'      : None,
			'workstation_name' : None,
			'ntlm_downgrade'   : False,
		},
		"Windows10_15063_channel" : {
			'flags'            :  NegotiateFlags.NEGOTIATE_KEY_EXCH|
								  NegotiateFlags.NEGOTIATE_128|
								  NegotiateFlags.NEGOTIATE_VERSION|
								  NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|
								  NegotiateFlags.NEGOTIATE_ALWAYS_SIGN|
								  NegotiateFlags.NEGOTIATE_NTLM|
								  NegotiateFlags.NEGOTIATE_SIGN|
								  NegotiateFlags.NEGOTIATE_SEAL|
								  NegotiateFlags.REQUEST_TARGET|
								  NegotiateFlags.NEGOTIATE_UNICODE,
			'version'          : Version.construct(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_10, minor = WindowsMinorVersion.WINDOWS_MINOR_VERSION_0, build = 15063 ),
			'domain_name'      : None,
			'workstation_name' : None,
			'ntlm_downgrade'   : False,
		},
		"Windows10_15063_old" : {
			'flags'            :  NegotiateFlags.NEGOTIATE_56|
								  NegotiateFlags.NEGOTIATE_KEY_EXCH|
								  NegotiateFlags.NEGOTIATE_128|
								  NegotiateFlags.NEGOTIATE_VERSION|
								  NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|
								  NegotiateFlags.NEGOTIATE_ALWAYS_SIGN|
								  NegotiateFlags.NEGOTIATE_NTLM|
								  NegotiateFlags.NEGOTIATE_LM_KEY|
								  NegotiateFlags.NEGOTIATE_SIGN|
								  NegotiateFlags.REQUEST_TARGET|
								  NegotiateFlags.NTLM_NEGOTIATE_OEM|
								  NegotiateFlags.NEGOTIATE_UNICODE,
			'version'          : Version.construct(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_10, minor = WindowsMinorVersion.WINDOWS_MINOR_VERSION_0, build = 15063 ),
			'domain_name'      : None,
			'workstation_name' : None,
			'ntlm_downgrade'   : False,
		},
		"Windows10_15063_knowkey" : {
			'flags'            :  NegotiateFlags.NEGOTIATE_56|
								  NegotiateFlags.NEGOTIATE_KEY_EXCH|
								  NegotiateFlags.NEGOTIATE_128|
								  NegotiateFlags.NEGOTIATE_VERSION|
								  NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|
								  NegotiateFlags.NEGOTIATE_ALWAYS_SIGN|
								  NegotiateFlags.NEGOTIATE_NTLM|
								  NegotiateFlags.NEGOTIATE_LM_KEY|
								  NegotiateFlags.NEGOTIATE_SIGN|
								  NegotiateFlags.REQUEST_TARGET|
								  NegotiateFlags.NTLM_NEGOTIATE_OEM|
								  NegotiateFlags.NEGOTIATE_UNICODE,
			'version'          : Version.construct(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_10, minor = WindowsMinorVersion.WINDOWS_MINOR_VERSION_0, build = 15063 ),
			'domain_name'      : None,
			'workstation_name' : None,
			'ntlm_downgrade'   : False,
			'session_key'      : b'A'*16,
		},
		"Windows10_15063_nosign" : {
			'flags'            :  NegotiateFlags.NEGOTIATE_56|
								  NegotiateFlags.NEGOTIATE_KEY_EXCH|
								  NegotiateFlags.NEGOTIATE_128|
								  NegotiateFlags.NEGOTIATE_VERSION|
								  NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|
								  NegotiateFlags.NEGOTIATE_NTLM|
								  NegotiateFlags.NEGOTIATE_LM_KEY|
								  NegotiateFlags.REQUEST_TARGET|
								  NegotiateFlags.NTLM_NEGOTIATE_OEM|
								  NegotiateFlags.NEGOTIATE_UNICODE,
			'version'          : Version.construct(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_10, minor = WindowsMinorVersion.WINDOWS_MINOR_VERSION_0, build = 15063 ),
			'domain_name'      : None,
			'workstation_name' : None,
			'ntlm_downgrade'   : False,
		},
}


import os
from asyauth.protocols.ntlm import logger

from asyauth.protocols.ntlm.client.native import NTLMClientNative
from asyauth.common.winapi.constants import ISC_REQ

from wsnet.agent.direct.auth import WSNETDirectAuth


class NTLMClientWSNETDirect:
	def __init__(self, credential):
		self._authid = os.urandom(4).hex()
		self.credential = credential
		self.sspi = None
		self.iterations = 0
		
		self.session_key = None
		self.ntlm_ctx = NTLMClientNative(self.credential)

	def setup(self):
		return
		
	@property
	def ntlmChallenge(self):
		return self.ntlm_ctx.ntlmChallenge
	
	def get_seq_number(self):
		return self.ntlm_ctx.seq_number
		
	def get_sealkey(self, mode = 'Client'):
		return self.ntlm_ctx.get_sealkey(mode = mode)
			
	def get_signkey(self, mode = 'Client'):
		return self.ntlm_ctx.get_signkey(mode = mode)
	
	async def encrypt(self, data, sequence_no):
		return await self.ntlm_ctx.encrypt(data, data, sequence_no)

	async def decrypt(self, data, sequence_no, direction='init', auth_data=None):
		return await self.ntlm_ctx.decrypt(data, sequence_no, direction=direction, auth_data=auth_data)

	async def sign(self, data, message_no, direction=None, reset_cipher = False):
		return await self.ntlm_ctx.sign(data, message_no, direction=direction, reset_cipher = reset_cipher)

	async def verify(self, data, signature):
		return await self.ntlm_ctx.verify(data, signature)
		
	def SEAL(self, signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt):
		return self.ntlm_ctx.SEAL(signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt)
		
	def SIGN(self, signingKey, message, seqNum, cipher_encrypt):
		return self.ntlm_ctx.SIGN(signingKey, message, seqNum, cipher_encrypt)
	
	def get_session_key(self):
		return self.session_key
		
	def get_extra_info(self):
		return self.ntlm_ctx.get_extra_info()
		
	def is_extended_security(self):
		return self.ntlm_ctx.is_extended_security()
	
	def signing_needed(self):
		return self.ntlm_ctx.signing_needed()
	
	def encryption_needed(self):
		return self.ntlm_ctx.encryption_needed()
		
	async def encrypt(self, data, message_no):
		return await self.ntlm_ctx.encrypt(data, message_no)
		
	async def decrypt(self, data, sequence_no, direction='init', auth_data=None):
		return await self.ntlm_ctx.decrypt(data, sequence_no, direction=direction, auth_data=auth_data)
	
	async def authenticate(self, authData, flags:ISC_REQ = None, cb_data = None, spn=None):
		try:
			if flags is None:
				flags = ISC_REQ.CONNECTION
			
			logger.debug('[WSNETDirect][NTLM][%s] spn: %s' % (self._authid, spn))
			logger.debug('[WSNETDirect][NTLM][%s] flags: %s' % (self._authid, flags))
			logger.debug('[WSNETDirect][NTLM][%s] cb_data: %s' % (self._authid, cb_data))
			
			if self.sspi is None:
				self.sspi = WSNETDirectAuth(self.credential.subprotocol.get_url())
				_, err = await self.sspi.connect()
				if err is not None:
					raise err
				
			if authData is None:
				logger.debug('[WSNETDirect][NTLM][%s] Initializing auth context, fetching Negotaite' % self._authid)
				status, ctxattr, negotiate_raw, err = await self.sspi.authenticate('NTLM', '', '', 3, flags.value, authdata = b'')
				if err is not None:
					raise err
				logger.debug('[WSNETDirect][NTLM][%s][Response] Status: %s' % (self._authid, status))
				logger.debug('[WSNETDirect][NTLM][%s][Response] Response flags: %s' % (self._authid, ctxattr))
				logger.debug('[WSNETDirect][NTLM][%s][Response] Negotiate: %s' % (self._authid, negotiate_raw))
				self.iterations += 1
				self.ntlm_ctx.load_negotiate(negotiate_raw)
				return negotiate_raw, True, None
			else:
				logger.debug('[WSNETDirect][NTLM][%s] Challenge: %s' % (self._authid, authData))
				self.ntlm_ctx.load_challenge(authData)
				logger.debug('[WSNETDirect][NTLM][%s] Sending Challenge, getting Authenticate...' % self._authid)
				status, ctxattr, authenticate_raw, err = await self.sspi.authenticate('NTLM', '', '', 3, flags.value, authdata = authData)
				if err is not None:
					raise err
				
				self.ntlm_ctx.load_authenticate(authenticate_raw)
				logger.debug('[WSNETDirect][NTLM][%s][Response] Status: %s' % (self._authid, status))
				logger.debug('[WSNETDirect][NTLM][%s][Response] Response flags: %s' % (self._authid, ctxattr))
				logger.debug('[WSNETDirect][NTLM][%s][Response] Authenticate: %s' % (self._authid, authenticate_raw))

				logger.debug('[WSNETDirect][NTLM][%s] Fetching SessionKey...' % self._authid)
				self.session_key, err = await self.sspi.get_sessionkey()
				if err is not None:
					raise err
				
				logger.debug('[WSNETDirect][NTLM][%s][Response] SessionKey: %s' % (self._authid, self.session_key.hex()))
				self.ntlm_ctx.load_sessionkey(self.get_session_key())
				
				logger.debug('[WSNETDirect][NTLM][%s][NTLMINFO] %s' % (self._authid, self.ntlm_ctx.get_extra_info()))

				await self.sspi.disconnect()
				return authenticate_raw, False, None
		except Exception as e:
			return None, None, e
		
	


from minikerberos.protocol.asn1_structs import AP_REQ, AP_REP
from minikerberos.protocol.encryption import Enctype, Key, _enctype_table
from asyauth.common.winapi.winsspi import WinSSPI, SSPIPackage
from asyauth.common.winapi.constants import ISC_REQ
from asyauth.protocols.kerberos.gssapi import get_gssapi
from asyauth.protocols.kerberos.gssapismb import get_gssapi as gssapi_smb


class KerberosClientSSPI:
	def __init__(self, credential):
		self.iterations = 0
		self.credential = credential
		self.ksspi = WinSSPI(SSPIPackage.KERBEROS)
		self.gssapi = None
		self.etype = None
		self.actual_ctx_flags = None

		self.seq_number = None
		self.session_key = None

	def get_seq_number(self):
		"""
		Fetches the starting sequence number. This is either zero or can be found in the authenticator field of the 
		AP_REQ structure. As windows uses a random seq number AND a subkey as well, we can't obtain it by decrypting the 
		AP_REQ structure. Insead under the hood we perform an encryption operation via EncryptMessage API which will 
		yield the start sequence number
		"""
		if self.seq_number is None:
			self.seq_number = self.ksspi.get_seq_number()
		return self.seq_number

	def signing_needed(self):
		return ISC_REQ.INTEGRITY in self.actual_ctx_flags
	
	def encryption_needed(self):
		return ISC_REQ.CONFIDENTIALITY in self.actual_ctx_flags
		
	async def encrypt(self, data, message_no):
		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction, auth_data=auth_data)
	
	def get_session_key(self):
		if self.session_key is None:
			self.session_key = self.ksspi.get_session_key()
		return self.session_key
	
	async def authenticate(self, authData = None, flags = ISC_REQ.CONNECTION, seq_number = 0, client_name = None, spn=None, cb_data = None):
		try:
			#authdata is only for api compatibility reasons
			if ISC_REQ.USE_DCE_STYLE in flags or ISC_REQ.MUTUAL_AUTH in flags:
				if self.iterations == 0:
					flags = ISC_REQ.CONFIDENTIALITY | \
							ISC_REQ.INTEGRITY | \
							ISC_REQ.MUTUAL_AUTH | \
							ISC_REQ.REPLAY_DETECT | \
							ISC_REQ.SEQUENCE_DETECT|\
							ISC_REQ.USE_DCE_STYLE
							
					token, self.actual_ctx_flags, err = self.ksspi.get_ticket_for_spn(spn, flags = flags, token_data = authData, cb_data = cb_data, client_name=client_name)
					if err is not None:
						raise err
					#print(token.hex())
					self.iterations += 1
					return token, True, None
				
				elif self.iterations == 1:
					flags = ISC_REQ.USE_DCE_STYLE		
					token, self.actual_ctx_flags, err = self.ksspi.get_ticket_for_spn(spn, flags = flags, token_data = authData, cb_data = cb_data, client_name=client_name)
					if err is not None:
						raise err
					
					aprep = AP_REP.load(token).native
					subkey = Key(aprep['enc-part']['etype'], self.get_session_key())

					self.get_seq_number()
					
					self.gssapi = gssapi_smb(subkey)
					
					self.iterations += 1
					return token, False, None
					
				else:
					raise Exception('SSPI Kerberos -RPC - auth encountered too many calls for authenticate.')
				
			else:
				apreqraw, self.actual_ctx_flags, err = self.ksspi.get_ticket_for_spn(spn, flags = flags, cb_data = cb_data, client_name=client_name)
				if err is not None:
					raise err
				
				if self.encryption_needed() is True or self.signing_needed() is True:
					self.get_seq_number()
					apreq = AP_REQ.load(apreqraw)
					skey = Key(apreq.native['ticket']['enc-part']['etype'], self.get_session_key())

					self.gssapi = get_gssapi(skey)
				return apreqraw, False, None
		except Exception as e:
			return None, True, e


from typing import Tuple
from asyauth.common.winapi.constants import ISC_REQ
from asyauth.common.credentials import UniCredential

class ClientAuthAlgo:
	def __init__(self, credential:UniCredential):
		pass

	#def is_guest(self):
	#	raise NotImplementedError()

	def get_seq_number(self) -> int:
		"""Returns the !INITIAL! sequence number"""
		raise NotImplementedError()
	
	def get_session_key(self) -> bytes:
		"""Returns session key bytes"""
		raise NotImplementedError()

	async def encrypt(self, data:bytes, sequence_no:int):
		"""Encryptes message"""
		raise NotImplementedError()

	async def decrypt(self, data:bytes, sequence_no:int, direction='init', auth_data=None) -> Tuple[bytes, Exception]:
		"""Decrypts message. Also performs integrity checking."""
		raise NotImplementedError()

	async def sign(self, data:bytes, message_no:int, direction:str=None, reset_cipher:bool = False) -> bytes:
		"""Singing outgoing messages. The reset_cipher parameter is needed for calculating mechListMIC."""
		raise NotImplementedError()

	async def verify(self, data:bytes, signature:bytes) -> bool:
		"""Verifying message signature"""
		raise NotImplementedError()
	
	def signing_needed(self) -> bool:
		"""Returns wether signing is negotiated for this client"""
		raise NotImplementedError()
	
	def encryption_needed(self) -> bool:
		"""Returns wether encryption is negotiated for this client"""
		raise NotImplementedError()
		
	def get_session_key(self) -> bytes:
		"""Returns the session key bytes"""
		raise NotImplementedError()

	async def authenticate(self, authData:bytes, flags:ISC_REQ = None, cb_data:bytes = None, spn:str=None): # -> Tuple[bytes, bool, Exception]
		raise NotImplementedError()


import base64
import platform
from urllib.parse import urlparse, parse_qs
from asyauth.utils.paramprocessor import str_one, int_one, bool_one
from asyauth.common.constants import asyauthSecret, asyauthProtocol, asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol, SubProtocolNative, SubProtocolSSPI

class UniCredential:
	def __init__(self, secret:str = None, username:str = None, domain:str = None, stype:asyauthSecret = asyauthSecret.NONE, protocol:asyauthProtocol = None, subprotocol:SubProtocol = SubProtocolNative()):
		self.domain = domain
		self.username = username
		self.secret = secret
		self.stype = stype
		self.protocol = protocol
		self.subprotocol = subprotocol

		if stype in [asyauthSecret.PASS, asyauthSecret.PW]:
			self.stype = asyauthSecret.PASSWORD
		elif stype == asyauthSecret.PWB64:
			self.stype = asyauthSecret.PASSWORD
			self.secret = base64.b64decode(self.secret).decode()
		elif stype == asyauthSecret.PWHEX:
			self.stype = asyauthSecret.PASSWORD
			self.secret = bytes.fromhex(self.secret).decode()
		elif stype == asyauthSecret.PWPROMPT:
			import getpass
			self.stype = asyauthSecret.PASSWORD
			self.secret = getpass.getpass('Enter password: ')
	
	def build_context(self):
		# override this function
		raise NotImplementedError()

	@staticmethod
	def get_url_params():
		return {
			'timeout' : int_one,
			'dns' : str_one,
			'dc' : str_one,
		}

	@staticmethod
	def from_url(connection_url):
		from asysocks.unicomm.common.target import UniTarget, UniProto

		secret = None
		username = None
		domain = None
		stype = asyauthSecret.NONE
		protocol = asyauthProtocol.NONE
		subprotocol = SubProtocolNative()
		url_e = urlparse(connection_url)
		schemes = url_e.scheme.upper().split('+')
		if len(schemes) == 1:
			try:
				protocol = asyauthProtocol(schemes)
			except:
				pass
		else:
			auth_tags = schemes[1].replace('-','_')
			try:
				protocol = asyauthProtocol(auth_tags)
			except:
				auth_tags = schemes[1].split('-')
				if len(auth_tags) > 1:
					try:
						spt = asyauthSubProtocol(auth_tags[0])
					except:
						protocol = asyauthProtocol(auth_tags[0])
						stype = asyauthSecret(auth_tags[1])
					else:
						protocol = asyauthProtocol(auth_tags[1])
						query = None
						if url_e.query is not None:
							query = parse_qs(url_e.query)
						subprotocol = SubProtocol.from_url_params(spt, query)
						
				else:
					try:
						spt = asyauthSubProtocol(auth_tags[0])
						protocol = asyauthProtocol.NTLM
					except:
						protocol = asyauthProtocol(auth_tags[0])
						
		if url_e.username is not None:
			if url_e.username.find('\\') != -1:
				domain, username = url_e.username.split('\\')
				if domain == '.':
					domain = None
			else:
				domain = None
				username = url_e.username
		
		secret = url_e.password
		credobj = None
		if protocol == asyauthProtocol.KERBEROS:
			from asyauth.common.credentials.kerberos import KerberosCredential
			credobj = KerberosCredential
		
		elif protocol in [asyauthProtocol.NTLM, asyauthProtocol.SICILY]:
			from asyauth.common.credentials.ntlm import NTLMCredential
			credobj = NTLMCredential

		extraparams = {}
		if credobj is not None:
			extraparams = credobj.get_url_params()

		paramstemplate = UniCredential.get_url_params()
		params = dict.fromkeys(UniCredential.get_url_params(),None)
		extra = dict.fromkeys(extraparams.keys(),None)
		proxy_present = False
		if url_e.query is not None:
			query = parse_qs(url_e.query)
			for k in query:
				if k.startswith('proxy') is True:
					proxy_present = True
				if k in params:
					params[k] = paramstemplate[k](query[k])
				if k in extraparams:
					extra[k] = extraparams[k](query[k])
		
		if protocol in [asyauthProtocol.NTLM, asyauthProtocol.SICILY]:
			res = credobj(
				secret,
				username, 
				domain,
				stype,
				subprotocol=subprotocol,
			)
			if protocol == asyauthProtocol.SICILY:
				res.protocol = asyauthProtocol.SICILY
			return res
		
		elif protocol == asyauthProtocol.KERBEROS:
			proxies = None
			if proxy_present is True:
				from asysocks.unicomm.common.proxy import UniProxyTarget
				proxies = UniProxyTarget.from_url_params(url_e.query, url_e.hostname,endpoint_port=88)

			target = None
			if extra['dc'] is not None:
				target = UniTarget(extra['dc'], 88, UniProto.CLIENT_TCP, proxies = proxies, dns=params['dns'], dc_ip=extra['dc'])

			etypes = extra['etype'] if extra['etype'] is not None else [23,17,18]

			return credobj(
				secret, 
				username, 
				domain, 
				stype, 
				target = target, 
				altname = extra['altname'], 
				altdomain = extra['altdomain'],
				certdata=extra['certdata'],
				keydata=extra['keydata'],
				etypes = etypes, 
				subprotocol = subprotocol
			)
		else:
			return UniCredential(
				secret,
				username,
				domain,
				stype,
				protocol,
				subprotocol
			)

	def __str__(self):
		import enum
		t = '==== UniCredential ====\r\n'
		for k in self.__dict__:
			val = self.__dict__[k]
			if isinstance(val, enum.IntFlag):
				val = val
			elif isinstance(val, enum.Enum):
				val = val.name
			
			t += '%s: %s\r\n' % (k, str(val))
			
		return t
	
	@staticmethod
	def get_sspi_ntlm():
		"""Returns an NTLM credential object matching the current user (Windows only)"""
		if platform.system() != 'Windows':
			raise Exception('This function only works on Windows systems!')
		from winacl.functions.highlevel import get_logon_info
		from asyauth.common.credentials.ntlm import NTLMCredential
		userinfo = get_logon_info()
		return NTLMCredential(
			None,
			userinfo['username'],
			userinfo['domain'],
			stype=asyauthSecret.NONE,
			subprotocol = SubProtocolSSPI()
		)
	
	@staticmethod
	def get_sspi_kerberos():
		"""Returns a Kerberos credential object matching the current user (Windows only)"""
		if platform.system() != 'Windows':
			raise Exception('This function only works on Windows systems!')
		from winacl.functions.highlevel import get_logon_info
		from asyauth.common.credentials.kerberos import KerberosCredential
		from asysocks.unicomm.common.target import UniTarget, UniProto
		userinfo = get_logon_info()
		dctarget = UniTarget(
			userinfo['logonserver'], 
			88, 
			UniProto.CLIENT_TCP, 
			dc_ip=userinfo['logonserver'],
			domain = userinfo['dnsdomainname'],
		)
		return KerberosCredential(
			None,
			userinfo['username'],
			userinfo['dnsdomainname'],
			stype=asyauthSecret.NONE,
			target=dctarget,
			subprotocol = SubProtocolSSPI()
		)

	@staticmethod
	def get_sspi(authtype:str):
		if platform.system() != 'Windows':
			raise Exception('This function only works on Windows systems!')
		if authtype.upper() == 'NTLM':
			credobj = UniCredential.get_sspi_ntlm()
		elif authtype.upper() == 'KERBEROS':
			credobj = UniCredential.get_sspi_kerberos()
		else:
			raise Exception('Only NTLM or KERBEROS auth types supported here!')
		return credobj

	@staticmethod
	def get_sspi_spnego(authtype:str):
		"""Returns a SPNEGO credential object matching the current user (Windows only)"""
		if platform.system() != 'Windows':
			raise Exception('This function only works on Windows systems!')
		from asyauth.common.credentials.spnego import SPNEGOCredential
		credobj = UniCredential.get_sspi(authtype)
		return SPNEGOCredential([credobj])
	
	@staticmethod
	def get_sspi_spnego(authtype:str):
		"""Returns a CREDSSP credential object matching the current user (Windows only)"""
		if platform.system() != 'Windows':
			raise Exception('This function only works on Windows systems!')
		from asyauth.common.credentials.credssp import CREDSSPCredential
		credobj = UniCredential.get_sspi(authtype)
		return CREDSSPCredential([credobj])
	
	@staticmethod
	def get_help(protocol:str = '', authprotos:str = '', extraparams:str = ''):
		"""Returns user help regarding the credential url format"""
		template = """
URL format:

	protocol+authproto-secrettype://[domain]\\username:secret@[ip|hostname]:[port]/?param1=value1&param2=value2

protocol: The protocol to use (see below)
	%s
authproto (protocol dependent): 
	%s
username: The username to authenticate with
secret: The secret to authenticate with (depends on secrettype)
domain: The domain of the user
secrettype: The type of the secret (see below)
	password/pw/pass: A plaintext password
	pwb64: A base64 encoded password
	pwprompt: Password will be prompted for on STDIN
	pwhex: A hex encoded password
	nt: NT hash
	==== Kerberos only ====
	rc4: RC4 key (Kerberos only)
	aes: AES key (any size, Kerberos only)
	aes128: AES128 key (Kerberos only)
	aes256: AES256 key (Kerberos only)
	ccache: ccache file name (only local directory)
	ccachehex: A hex encoded ccache file
	ccacheb64: A base64 encoded ccache file
	keytab: keytab file name (only local directory)
	keytabhex: A hex encoded keytab file
	keytabb64: A base64 encoded keytab file
	pfx: pfx file name (only local directory)
	pfxhex: A hex encoded pfx file
	pfxb64: A base64 encoded pfx file
	pem: pem file name (only local directory)
	pemhex: A hex encoded pem file
	pemb64: A base64 encoded pem file
	certstore: Use Windows certificate store (Windows only, Kerberos only)
	kirbi: kirbi file name (only local directory)
	kirbihex: A hex encoded kirbi file
	kirbib64: A base64 encoded kirbi file
	==== SSH Only ====
	sshprivkey: ssh private key file name (only local directory)
	sshprivkeystr: ssh private key file data in string format
	sshprivkeyb64: A base64 encoded ssh private key file

Extra parameters are in the form of param=value. The following parameters are supported:
	altname: Alternative name for certificate authentication
	altdomain: Alternative domain for certificate authentication
	etype: Supported encryption type to use (Kerberos only)
	dc: The domain controller to use (Kerberos only)
	dns: Don't use this.
	%s
"""
		return template % (protocol, authprotos, extraparams)
		#add examples



from asyauth.common.credentials.credssp import CREDSSPCredential
from asyauth.common.credentials.kerberos import KerberosCredential
from asyauth.common.credentials.ntlm import NTLMCredential
from asyauth.common.credentials.spnego import SPNEGOCredential
from asyauth.common.credentials.spnegoex import SPNEGOEXCredential


__all__ = [
	'CREDSSPCredential', 
	'KerberosCredential', 
	'NTLMCredential', 
	'SPNEGOCredential',
	'SPNEGOEXCredential',
]


from asyauth.common.winapi.winsspi import WinSSPI, SSPIPackage
from asyauth.common.winapi.constants import ISC_REQ, SSPIResult
from asyauth.protocols.ntlm.client.native import NTLMClientNative
from asyauth.common.credentials.ntlm import NTLMCredential

class NTLMClientSSPI:
	def __init__(self, credential:NTLMCredential):
		self.credential = credential
		self.sspi = WinSSPI(SSPIPackage.NTLM)
		self.session_key = None
		self.ntlm_ctx = NTLMClientNative(self.credential)
		self.flags = ISC_REQ.CONNECTION

	@property
	def ntlmChallenge(self):
		return self.ntlm_ctx.ntlmChallenge
	
	def get_seq_number(self):
		return self.ntlm_ctx.seq_number
		
	def get_sealkey(self, mode = 'Client'):
		return self.ntlm_ctx.get_sealkey(mode = mode)
			
	def get_signkey(self, mode = 'Client'):
		return self.ntlm_ctx.get_signkey(mode = mode)
	
	async def encrypt(self, data, sequence_no):
		return await self.ntlm_ctx.encrypt(data, data, sequence_no)

	async def decrypt(self, data, sequence_no, direction='init', auth_data=None):
		return await self.ntlm_ctx.decrypt(data, sequence_no, direction=direction, auth_data=auth_data)

	async def sign(self, data, message_no, direction=None, reset_cipher = False):
		return await self.ntlm_ctx.sign(data, message_no, direction=direction, reset_cipher = reset_cipher)

	async def verify(self, data, signature):
		return await self.ntlm_ctx.verify(data, signature)
		
	def SEAL(self, signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt):
		return self.ntlm_ctx.SEAL(signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt)
		
	def SIGN(self, signingKey, message, seqNum, cipher_encrypt):
		return self.ntlm_ctx.SIGN(signingKey, message, seqNum, cipher_encrypt)
	
	def get_session_key(self):
		if not self.session_key:
			self.session_key = self.sspi.get_session_key()
		
		return self.session_key
		
	def get_extra_info(self):
		return self.ntlm_ctx.get_extra_info()
		
	def is_extended_security(self):
		return self.ntlm_ctx.is_extended_security()
	
	def signing_needed(self):
		return self.ntlm_ctx.signing_needed()
	
	def encryption_needed(self):
		return self.ntlm_ctx.encryption_needed()
		
	async def encrypt(self, data, message_no):
		return await self.ntlm_ctx.encrypt(data, message_no)
		
	async def decrypt(self, data, sequence_no, direction='init', auth_data=None):
		return await self.ntlm_ctx.decrypt(data, sequence_no, direction=direction, auth_data=auth_data)
	
	async def authenticate(self, authData = None, flags = None, seq_number = 0, spn = None, cb_data = None):
		to_continue = False
		if flags is None:
			flags = self.flags
		
		if authData is None:
			self.sspi.acquire_handle(None, spn) #self.credential.username
			result, data, self.flags, expiry = self.sspi.initialize_context(spn, authData, flags=flags, cb_data=cb_data)
			if result == SSPIResult.CONTINUE:
				to_continue = True
			self.ntlm_ctx.load_negotiate(data)
			return data, to_continue, None
		else:
			self.ntlm_ctx.load_challenge(authData)
			result, data, self.flags, expiry = self.sspi.initialize_context(spn, authData, flags=flags, cb_data=cb_data)
			if result == SSPIResult.CONTINUE:
				to_continue = True
		
			self.ntlm_ctx.load_authenticate(data)
			self.ntlm_ctx.load_sessionkey(self.get_session_key())
			
			return data, to_continue, None
			
	


from asyauth import logger
from asyauth.protocols.ntlm.client.native import NTLMClientNative
from wsnet.operator.sspiproxy import WSNETSSPIProxy
from asyauth.common.winapi.constants import ISC_REQ

class NTLMClientSSPIProxy:
	def __init__(self, credential):
		self.credential = credential
		url = '%s://%s:%s' % (self.settings.proto, self.settings.host, self.settings.port)
		self.sspi = WSNETSSPIProxy(url, self.settings.agent_id)
		self.iterations = 0
		
		self.session_key = None
		self.ntlm_ctx = NTLMClientNative(self.credential)

	def setup(self):
		return
		
	@property
	def ntlmChallenge(self):
		return self.ntlm_ctx.ntlmChallenge
		
	def get_sealkey(self, mode = 'Client'):
		return self.ntlm_ctx.get_sealkey(mode = mode)
			
	def get_signkey(self, mode = 'Client'):
		return self.ntlm_ctx.get_signkey(mode = mode)
		
		
	def SEAL(self, signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt):
		return self.ntlm_ctx.SEAL(signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt)
		
	def SIGN(self, signingKey, message, seqNum, cipher_encrypt):
		return self.ntlm_ctx.SIGN(signingKey, message, seqNum, cipher_encrypt)
	
	def get_session_key(self):
		return self.session_key
		
	def get_extra_info(self):
		return self.ntlm_ctx.get_extra_info()
		
	def is_extended_security(self):
		return self.ntlm_ctx.is_extended_security()
	
	async def authenticate(self, authData = b'', flags = None, seq_number = 0, is_rpc = False):
		try:
			if is_rpc is True and flags is None:
				flags = ISC_REQ.REPLAY_DETECT | ISC_REQ.CONFIDENTIALITY| ISC_REQ.USE_SESSION_KEY| ISC_REQ.INTEGRITY| ISC_REQ.SEQUENCE_DETECT| ISC_REQ.CONNECTION
			elif flags is None:
				flags = ISC_REQ.CONNECTION

			if authData is None:
				status, ctxattr, data, err = await self.sspi.authenticate('NTLM', '', '', 3, flags.value, authdata = b'')
				if err is not None:
					raise err
				self.iterations += 1
				self.ntlm_ctx.load_negotiate(data)
				return data, True, None
			else:
				self.ntlm_ctx.load_challenge(authData)
				status, ctxattr, data, err = await self.sspi.authenticate('NTLM', '', '', 3, flags.value, authdata = authData)
				if err is not None:
					raise err
				if err is None:
					self.ntlm_ctx.load_authenticate(data)
					self.session_key, err = await self.sspi.get_sessionkey()
					if err is not None:
						raise err
					self.ntlm_ctx.load_sessionkey(self.get_session_key())
				
				await self.sspi.disconnect()
				return data, False, None
		except Exception as e:
			return None, None, e
		
	


from asyauth.common.constants import asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol


class SubProtocolCustom(SubProtocol):
	def __init__(self, factoryobj):
		SubProtocol.__init__(self, asyauthSubProtocol.CUSTOM)
		self.factoryobj = factoryobj

	@staticmethod
	def from_url_params(query_params = None):
		raise NotImplementedError()
	
	def build_context(self, credential):
		return self.factoryobj.build_context(credential)
	
	def __deepcopy__(self, memo=None):
		return SubProtocolCustom(self.factoryobj)


import enum
import io
import os

from minikerberos.protocol.constants import EncryptionType
from minikerberos.protocol import encryption
from unicrypto.hashlib import md5
from unicrypto import hmac
from unicrypto.symmetric import RC4


# https://tools.ietf.org/html/draft-raeburn-krb-rijndael-krb-05
# https://tools.ietf.org/html/rfc2478
# https://tools.ietf.org/html/draft-ietf-krb-wg-gssapi-cfx-02

GSS_WRAP_HEADER = b'\x60\x2b\x06\x09\x2a\x86\x48\x86\xf7\x12\x01\x02\x02'

class GSSAPIFlags(enum.IntFlag):
	GSS_C_DCE_STYLE     = 0x1000
	GSS_C_DELEG_FLAG    = 1
	GSS_C_MUTUAL_FLAG   = 2
	GSS_C_REPLAY_FLAG   = 4
	GSS_C_SEQUENCE_FLAG = 8
	GSS_C_CONF_FLAG     = 0x10
	GSS_C_INTEG_FLAG    = 0x20
	
class KG_USAGE(enum.Enum):
	ACCEPTOR_SEAL  = 22
	ACCEPTOR_SIGN  = 23
	INITIATOR_SEAL = 24
	INITIATOR_SIGN = 25
	
class FlagsField(enum.IntFlag):
	SentByAcceptor = 0
	Sealed = 2
	AcceptorSubkey = 4
	
# https://tools.ietf.org/html/rfc4757 (7.2)
class GSSMIC_RC4:
	def __init__(self):
		self.TOK_ID = b'\x01\x01'
		self.SGN_ALG = b'\x11\x00' #HMAC
		self.Filler = b'\xff'*4
		self.SND_SEQ = None
		self.SGN_CKSUM = None

	@staticmethod
	def from_bytes(data):
		return GSSMIC_RC4.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		mic = GSSMIC_RC4()
		mic.TOK_ID = buff.read(2)
		mic.SGN_ALG = buff.read(2)
		mic.Filler =  buff.read(4)
		mic.SND_SEQ = buff.read(8)
		mic.SGN_CKSUM = buff.read(8)
		
		return mic
		
	def to_bytes(self):
		t  = self.TOK_ID
		t += self.SGN_ALG
		t += self.Filler
		t += self.SND_SEQ
		if self.SGN_CKSUM is not None:
			t += self.SGN_CKSUM
	
		return t
		
class GSSWRAP_RC4:
	def __init__(self):
		self.TOK_ID = b'\x02\x01'
		self.SGN_ALG = b'\x11\x00' #HMAC
		self.SEAL_ALG = None
		self.Filler = b'\xFF' * 2
		self.SND_SEQ = None
		self.SGN_CKSUM = None
		self.Confounder = None
	
	@staticmethod
	def from_bytes(data):
		return GSSWRAP_RC4.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		wrap = GSSWRAP_RC4()
		wrap.TOK_ID = buff.read(2)
		wrap.SGN_ALG = buff.read(2)
		wrap.SEAL_ALG = buff.read(2)
		wrap.Filler = buff.read(2)
		wrap.SND_SEQ = buff.read(8)
		wrap.SGN_CKSUM = buff.read(8)
		wrap.Confounder = buff.read(8)
		
		return wrap
	
	def to_bytes(self):
		t  = self.TOK_ID
		t += self.SGN_ALG
		t += self.SEAL_ALG
		t += self.Filler
		t += self.SND_SEQ
		
		if self.SGN_CKSUM:
			t += self.SGN_CKSUM
			if self.Confounder:
				t += self.Confounder
		
	
		return t
		
class GSSAPI_RC4:
	def __init__(self, session_key):
		self.session_key = session_key
	
	def GSS_GetMIC(self, data, sequenceNumber, direction = 'init'):
		GSS_GETMIC_HEADER = b'\x60\x23\x06\x09\x2a\x86\x48\x86\xf7\x12\x01\x02\x02'
		
		# Let's pad the data
		pad = (4 - (len(data) % 4)) & 0x3
		padStr = bytes([pad]) * pad
		data += padStr
		
		mic = GSSMIC_RC4()
		
		if direction == 'init':
			mic.SND_SEQ = sequenceNumber.to_bytes(4, 'big', signed = False) + b'\x00'*4
		else:
			mic.SND_SEQ = sequenceNumber.to_bytes(4, 'big', signed = False) + b'\xff'*4
		
		Ksign_ctx = hmac.new(self.session_key.contents, digestmod='md5')
		Ksign_ctx.update(b'signaturekey\0')
		Ksign = Ksign_ctx.digest()
		
		id = 15
		temp = md5( id.to_bytes(4, 'little', signed = False) +  mic.to_bytes()[:8] ).digest()
		chksum_ctx = hmac.new(Ksign, digestmod='md5')
		chksum_ctx.update(temp)
		mic.SGN_CKSUM = chksum_ctx.digest()[:8]
		
		id = 0
		temp = hmac.new(self.session_key.contents, digestmod='md5')
		temp.update(id.to_bytes(4, 'little', signed = False))
		
		Kseq_ctx = hmac.new(temp.digest(), digestmod='md5')
		Kseq_ctx.update(mic.SGN_CKSUM)
		Kseq = Kseq_ctx.digest()
		
		mic.SGN_CKSUM = RC4(Kseq).encrypt(mic.SND_SEQ)
		
		return GSS_GETMIC_HEADER + mic.to_bytes()
		
	
	def GSS_Wrap(self, data, seq_num, direction = 'init', encrypt=True, auth_data=None):
		GSS_WRAP_HEADER = b'\x60\x2b\x06\x09\x2a\x86\x48\x86\xf7\x12\x01\x02\x02'
		
		pad = (8 - (len(data) % 8)) & 0x7
		padStr = bytes([pad]) * pad
		data += padStr
		
		token = GSSWRAP_RC4()
		token.SEAL_ALG = b'\x10\x00'
		
		if direction == 'init':
			token.SND_SEQ = seq_num.to_bytes(4, 'big', signed = False) + b'\x00'*4
		else:
			token.SND_SEQ = seq_num.to_bytes(4, 'big', signed = False) + b'\xff'*4
			
		token.Confounder = os.urandom(8)
		
		temp = hmac.new(self.session_key.contents, digestmod='md5')
		temp.update(b'signaturekey\0')
		Ksign = temp.digest()
		
		id = 13
		Sgn_Cksum = md5( id.to_bytes(4, 'little', signed = False) +  token.to_bytes()[:8] + token.Confounder + data).digest()
		temp = hmac.new(Ksign, digestmod='md5')
		temp.update(Sgn_Cksum)
		token.SGN_CKSUM = temp.digest()[:8]
		
		
		klocal = b''
		for b in self.session_key .contents:
			klocal += bytes([b ^ 0xf0])
			
		id = 0
		temp = hmac.new(klocal, digestmod='md5')
		temp.update(id.to_bytes(4, 'little', signed = False))
		temp = hmac.new(temp.digest(), digestmod='md5')
		temp.update(seq_num.to_bytes(4, 'big', signed = False))
		Kcrypt = temp.digest()
		
		id = 0
		temp = hmac.new(self.session_key.contents, digestmod='md5')
		temp.update(id.to_bytes(4, 'little', signed = False))
		temp = hmac.new(temp.digest(), digestmod='md5')
		temp.update(token.SGN_CKSUM)
		Kseq = temp.digest()
		
		token.SND_SEQ = RC4(Kseq).encrypt(token.SND_SEQ)
		
		
		if auth_data is not None:
			wrap = GSSWRAP_RC4.from_bytes(auth_data[8 + len(GSS_WRAP_HEADER):])
			
			id = 0
			temp = hmac.new(self.session_key.contents, digestmod='md5')
			temp.update(id.to_bytes(4, 'little', signed = False))
			temp = hmac.new(temp.digest(), digestmod='md5')
			temp.update(wrap.SGN_CKSUM)
			
			snd_seq = RC4(temp.digest()).encrypt(wrap.SND_SEQ)
			
			id = 0
			temp = hmac.new(klocal, digestmod='md5')
			temp.update(id.to_bytes(4, 'little', signed = False))
			temp = hmac.new(temp.digest(), digestmod='md5')
			temp.update(snd_seq[:4])
			Kcrypt = temp.digest()
			
			rc4 = RC4(Kcrypt)
			cipherText = rc4.decrypt(token.Confounder + data)[8:]
			
		elif encrypt is True:
			rc4 = RC4(Kcrypt)
			token.Confounder = rc4.encrypt(token.Confounder)
			cipherText = rc4.encrypt(data)
		
		else:
			cipherText = data
			
		finalData = GSS_WRAP_HEADER + token.to_bytes()
		return cipherText, finalData
		

	def GSS_Unwrap(self, data, seq_num, direction='init', auth_data = None):
		return self.GSS_Wrap(data, seq_num, direction, False, auth_data)
	
# 4.2.6.1. MIC Tokens
class GSSMIC:
	def __init__(self):
		self.TOK_ID = b'\x04\x04'
		self.Flags = None
		self.Filler = b'\xFF' * 5
		self.SND_SEQ = None
		self.SGN_CKSUM = None
		
	@staticmethod
	def from_bytes(data):
		return GSSMIC.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		m = GSSMIC()
		m.TOK_ID = buff.read(2)
		m.Flags = FlagsField(int.from_bytes(buff.read(1), 'big', signed = False))
		m.Filler = buff.read(5)
		m.SND_SEQ = int.from_bytes(buff.read(8), 'big', signed = False)
		m.SGN_CKSUM = buff.read() #should know the size based on the algo!
		return m
		
	def to_bytes(self):
		t  = self.TOK_ID
		t += self.Flags.to_bytes(1, 'big', signed = False)
		t += self.Filler
		t += self.SND_SEQ.to_bytes(8, 'big', signed = False)
		if self.SGN_CKSUM is not None:
			t += self.SGN_CKSUM
		
		return t
		
# 4.2.6.2. Wrap Tokens
class GSSWrapToken:
	def __init__(self):
		self.TOK_ID = b'\x05\x04'
		self.Flags = None
		self.Filler = b'\xFF'
		self.EC = None
		self.RRC = None
		self.SND_SEQ = None
		self.Data = None
		
	@staticmethod
	def from_bytes(data):
		return GSSWrapToken.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		m = GSSWrapToken()
		m.TOK_ID = buff.read(2)
		m.Flags = FlagsField(int.from_bytes(buff.read(1), 'big', signed = False))
		m.Filler = buff.read(1)
		m.EC = int.from_bytes(buff.read(2), 'big', signed = False)
		m.RRC = int.from_bytes(buff.read(2), 'big', signed = False)
		m.SND_SEQ = int.from_bytes(buff.read(8), 'big', signed = False)
		return m
		
	def to_bytes(self):
		t  = self.TOK_ID
		t += self.Flags.to_bytes(1, 'big', signed = False)
		t += self.Filler
		t += self.EC.to_bytes(2, 'big', signed = False)
		t += self.RRC.to_bytes(2, 'big', signed = False)
		t += self.SND_SEQ.to_bytes(8, 'big', signed = False)
		if self.Data is not None:
			t += self.Data
		
		return t
		
class GSSAPI_AES:
	def __init__(self, session_key, cipher_type, checksum_profile):
		self.session_key = session_key
		self.checksum_profile = checksum_profile
		self.cipher_type = cipher_type
		self.cipher = None
		
	def rotate(self, data, numBytes):
		numBytes %= len(data)
		left = len(data) - numBytes
		result = data[left:] + data[:left]
		return result
		
	def unrotate(self, data, numBytes):
		numBytes %= len(data)
		result = data[numBytes:] + data[:numBytes]
		return result
		
	def GSS_GetMIC(self, data, seq_num):
		pad = (4 - (len(data) % 4)) & 0x3
		padStr = bytes([pad]) * pad
		data += padStr
		
		m = GSSMIC()
		m.Flags = FlagsField.AcceptorSubkey
		m.SND_SEQ = seq_num
		checksum_profile = self.checksum_profile()
		m.checksum = checksum_profile.checksum(self.session_key, KG_USAGE.INITIATOR_SIGN.value, data + m.to_bytes()[:16])
		
		return m.to_bytes()
		
	def GSS_Wrap(self, data, seq_num):
		cipher = self.cipher_type()
		pad = (cipher.blocksize - (len(data) % cipher.blocksize)) & 15
		padStr = b'\xFF' * pad
		data += padStr
		
		t = GSSWrapToken()
		t.Flags = FlagsField.Sealed | FlagsField.AcceptorSubkey
		t.EC = pad
		t.RRC = 0
		t.SND_SEQ = seq_num
		
		cipher_text = cipher.encrypt(self.session_key, KG_USAGE.INITIATOR_SEAL.value,  data + t.to_bytes(), None)
		t.RRC = 28 #[RFC4121] section 4.2.5
		cipher_text = self.rotate(cipher_text, t.RRC + t.EC)
		
		ret1 = cipher_text[16 + t.RRC + t.EC:]
		ret2 = t.to_bytes() + cipher_text[:16 + t.RRC + t.EC]
		
		return ret1, ret2
		
	def GSS_Unwrap(self, data, seq_num, direction='init', auth_data = None):
		cipher = self.cipher_type()
		t = GSSWrapToken.from_bytes(auth_data[8:])
		rotated = auth_data[16+8:] + data
		
		
		cipher_text = self.unrotate(rotated, t.RRC + t.EC)
		plain_text = cipher.decrypt(self.session_key, KG_USAGE.ACCEPTOR_SEAL.value, cipher_text)
		
		return plain_text[:-(t.EC + 16)], None
		
def get_gssapi(session_key):
	if session_key.enctype == encryption.Enctype.AES256:
		return GSSAPI_AES(session_key, encryption._AES256CTS, encryption._SHA1AES256)
	if session_key.enctype == encryption.Enctype.AES128:
		return GSSAPI_AES(session_key, encryption._AES128CTS, encryption._SHA1AES128)
	elif session_key.enctype == encryption.Enctype.RC4:
		return GSSAPI_RC4(session_key)
	else:
		raise Exception('Unsupported etype %s' % session_key.enctype)
		
		
def test():
	data_padded= bytes.fromhex('810e00001a204de2d64fd111a3da0000f875ae0d1c4500003400000034000000008040050000000000000000000000000000000000000000000000000000000000000000000000000000000000000000ffffffffffffffffffffffffffffffff')
	token_1= bytes.fromhex('050406ff000c00000000000000000000')
	cipherText_1 = bytes.fromhex('0880ed78d6196dde3f3fb23eeea650bc4ae025fa2a9c337c75c024d9d8f0186c75a4a9060e2a40a9ad024317bf5df6a86cb4a764a9ca36843f8fa4f99c03e2bde46f5a29aafc83dacdf9f0a5677446b5d910417142dc7b7ba7ded76cddc4acf9bf7ed44008cb9850e5701f2f9285dad6463ca8d0e365d4f1700f3d054e242ebcde2f3146ddd411a627af7486')
	cipherText_2 = bytes.fromhex('08cb9850e5701f2f9285dad6463ca8d0e365d4f1700f3d054e242ebcde2f3146ddd411a627af74860880ed78d6196dde3f3fb23eeea650bc4ae025fa2a9c337c75c024d9d8f0186c75a4a9060e2a40a9ad024317bf5df6a86cb4a764a9ca36843f8fa4f99c03e2bde46f5a29aafc83dacdf9f0a5677446b5d910417142dc7b7ba7ded76cddc4acf9bf7ed440')



	session_key = encryption.Key( encryption.Enctype.AES256 , bytes.fromhex('3e242e91996aadd513ecb1bc2369e44183e08e08c51550fa4b681e77f75ed8e1'))
	data = bytes.fromhex('810e00001a204de2d64fd111a3da0000f875ae0d1c4500003400000034000000008040050000000000000000000000000000000000000000000000000000000000000000000000000000000000000000ffffffff')
	sequenceNumber = 0
	ret1 = bytes.fromhex('4ae025fa2a9c337c75c024d9d8f0186c75a4a9060e2a40a9ad024317bf5df6a86cb4a764a9ca36843f8fa4f99c03e2bde46f5a29aafc83dacdf9f0a5677446b5d910417142dc7b7ba7ded76cddc4acf9bf7ed440')
	ret2 = bytes.fromhex('050406ff000c001c000000000000000008cb9850e5701f2f9285dad6463ca8d0e365d4f1700f3d054e242ebcde2f3146ddd411a627af74860880ed78d6196dde3f3fb23eeea650bc')
	
	gssapi = get_gssapi(session_key)
	r1, r2 = gssapi.GSS_Wrap(data, sequenceNumber)
	
	gssapi.GSS_Unwrap(r1, 0, auth_data = b'\xff'*8 + r2)
	
	print(r1.hex())
	print(ret1.hex())
	
	assert r1 == ret1
	assert r2 == ret2

if __name__ == '__main__':
	test()



from asyauth.common.credentials import UniCredential
from asyauth.common.constants import asyauthSecret, asyauthProtocol, asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol
from asyauth.common.subprotocols import SubProtocolNative

from asyauth.protocols.ntlm.structures.version import Version, WindowsMajorVersion, WindowsMinorVersion
from asyauth.protocols.ntlm.structures.negotiate_flags import NegotiateFlags

ASYAUTH_NTLMCRED_SUPPORTED_STYPE = [
	asyauthSecret.PASSWORD,
	asyauthSecret.PASS,
	asyauthSecret.PW,
	asyauthSecret.PWHEX,
	asyauthSecret.NT,
	asyauthSecret.RC4
]
class NTLMCredential(UniCredential):
	def __init__(self, secret, username, domain, stype:asyauthSecret, subprotocol:SubProtocol = SubProtocolNative()):
		UniCredential.__init__(
			self, 
			secret = secret,
			username = username,
			domain = domain,
			stype = stype,
			protocol = asyauthProtocol.NTLM,
			subprotocol=subprotocol
		)

		if self.stype is asyauthSecret.RC4:
			self.stype = asyauthSecret.NT
		if isinstance(self.subprotocol, SubProtocolNative) and self.stype not in ASYAUTH_NTLMCRED_SUPPORTED_STYPE:
			raise Exception('Unsupported Secret Type for NTLM auth: %s' % self.stype)
		
		self.is_guest = False
		self.negotiate_workstation = None
		self.negotiate_domain = None
		self.ntlm_version = 2
		self.flags = NegotiateFlags.NEGOTIATE_KEY_EXCH|\
			NegotiateFlags.NEGOTIATE_128|\
			NegotiateFlags.NEGOTIATE_VERSION|\
			NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|\
			NegotiateFlags.NEGOTIATE_ALWAYS_SIGN|\
			NegotiateFlags.NEGOTIATE_NTLM|\
			NegotiateFlags.NEGOTIATE_SIGN|\
			NegotiateFlags.NEGOTIATE_SEAL|\
			NegotiateFlags.REQUEST_TARGET|\
			NegotiateFlags.NEGOTIATE_UNICODE
			
		self.negotiate_version = Version.construct(
			WindowsMajorVersion.WINDOWS_MAJOR_VERSION_10, 
			minor = WindowsMinorVersion.WINDOWS_MINOR_VERSION_0, 
			build = 15063 
		)

		### these fields are for testing!
		self.challenge = None # none -> it will generate a random challenge for the context
		self.timestamp = None
		self.session_key = None
			
	@staticmethod
	def get_url_params():
		return {

		}
	
	def build_context(self):
		if self.subprotocol.type == asyauthSubProtocol.NATIVE:
			from asyauth.protocols.ntlm.client.native import NTLMClientNative
			return NTLMClientNative(self)

		elif self.subprotocol.type == asyauthSubProtocol.SSPI:
			from asyauth.protocols.ntlm.client.sspi import NTLMClientSSPI
			return NTLMClientSSPI(self)

		elif self.subprotocol.type == asyauthSubProtocol.SSPIPROXY:
			from asyauth.protocols.ntlm.client.sspiproxy import NTLMClientSSPIProxy
			return NTLMClientSSPIProxy(self)

		elif self.subprotocol.type == asyauthSubProtocol.WSNET:
			from asyauth.protocols.ntlm.client.wsnet import NTLMClientWSNET
			return NTLMClientWSNET(self)
		
		elif self.subprotocol.type == asyauthSubProtocol.WSNETDIRECT:
			from asyauth.protocols.ntlm.client.wsnetdirect import NTLMClientWSNETDirect
			return NTLMClientWSNETDirect(self)
		
		elif self.subprotocol.type == asyauthSubProtocol.CUSTOM:
			return self.subprotocol.factoryobj.build_context(self)
		else:
			raise Exception('Unsupported subprotocol "%s"' % self.subprotocol)
	
	def __str__(self):
		import enum
		t = '==== NTLMCredential ====\r\n'
		for k in self.__dict__:
			val = self.__dict__[k]
			if isinstance(val, enum.IntFlag):
				val = val
			elif isinstance(val, enum.Enum):
				val = val.name
			
			t += '%s: %s\r\n' % (k, str(val))
			
		return t


from setuptools import setup, find_packages
import re

VERSIONFILE="asyauth/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))


setup(
	# Application name:
	name="asyauth",

	# Version number (initial):
	version=verstr,

	# Application author details:
	author="Tamas Jos",
	author_email="info@skelsecprojects.com",

	# Packages
	packages=find_packages(exclude=["tests*"]),

	# Include additional files into the package
	include_package_data=True,

	# Details
	url="https://github.com/skelsec/asyauth",

	zip_safe = False,
	#
	# license="LICENSE.txt",
	description="Unified authentication library",
	long_description="Unified authentication library",

	# long_description=open("README.txt").read(),
	python_requires='>=3.7',
	classifiers=[
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	install_requires=[
		'unicrypto>=0.0.10',
		'asn1crypto>=1.3.0',
		'asysocks>=0.2.7',
		'minikerberos>=0.4.2',
	],
)


import platform
from asyauth.common.constants import asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol


class SubProtocolSSPI(SubProtocol):
	def __init__(self):
		SubProtocol.__init__(self, asyauthSubProtocol.SSPI)
		if platform.system() != 'Windows':
			raise Exception("SSPI subprotocol only works on Windows!")

	@staticmethod
	def from_url_params(query_params = None):
		return SubProtocolSSPI()


# This is working, but also a work in progress!!!

from asyauth.protocols.spnego.messages.asn1_structs import *
import asyncio
from typing import Callable

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-spng/d4f2b41c-5f9e-4e11-98d0-ade76467095d


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-spng/94ccc4f8-d224-495f-8d31-4f58d1af598e
## SPNEGO has been assigned the following object identifier (OID): so.org.dod.internet.security.mechanism.snego (1.3.6.1.5.5.2)

class SPNEGORelay:
	def __init__(self, auth_relay_queue:asyncio.Queue):
		self.auth_relay_queue = auth_relay_queue
		self.authentication_contexts = {}
		self.selected_authentication_context = None
		self.selected_authentication_context_server = None
		self.selected_mechtype = None
		self.selected_mechtype_server = None
		self.iteration_ctr = 0
		self.start_client_evt = None
		self.iteration_ctr_server = 0
		self.mic_data = None
		self.__server_side_crash = None
		self.__client_side_crash = None
		self.__server_latest_token = None
		self.__relay_queue_notified = False
		self.__authtype = None
	
	def setup(self, log_q = None):
		for k in self.authentication_contexts:
			self.authentication_contexts[k].setup(log_q)
			self.start_client_evt = self.authentication_contexts[k].start_client_evt
	
	async def notify_relay(self, authtype = None):
		if self.__relay_queue_notified is False:
			self.__authtype = authtype
			await self.auth_relay_queue.put(self)
			self.__relay_queue_notified = True

	def is_guest(self):
		if self.selected_authentication_context is None:
			raise Exception('Call this after selecting auth method!')
		
		if hasattr(self.selected_authentication_context, 'is_guest') is True:
			return self.selected_authentication_context.is_guest()
		return False

	async def sign(self, data, message_no, direction='init'):
		return await self.selected_authentication_context.sign(data, message_no, direction=direction)
		
	async def encrypt(self, data, message_no):
		return await self.selected_authentication_context.encrypt(data, message_no)

	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return await self.selected_authentication_context.decrypt(data, message_no, direction=direction, auth_data=auth_data)
	
	def get_ntlm(self):
		if 'NTLMSSP - Microsoft NTLM Security Support Provider' in self.authentication_contexts:
			return self.authentication_contexts['NTLMSSP - Microsoft NTLM Security Support Provider']
		return None
		
	def add_auth_context(self, name, ctx):
		"""
		Add an authentication context to the given authentication context name.
		Valid names are:
			'NTLMSSP - Microsoft NTLM Security Support Provider'
			'MS KRB5 - Microsoft Kerberos 5'
			'KRB5 - Kerberos 5'
			'KRB5 - Kerberos 5 - User to User'
			'NEGOEX - SPNEGO Extended Negotiation Security Mechanism'
			
		Context MUST be already set up!
		"""
		self.authentication_contexts[name] = ctx
		
	def select_common_athentication_type(self, mech_types):
		for auth_type_name in self.authentication_contexts:
			if auth_type_name in mech_types:
				return auth_type_name, self.authentication_contexts[auth_type_name]
				
		return None, None
	
	async def process_ctx_authenticate_server(self, token_data, include_negstate = False, flags = None, seq_number = 0, is_rpc = False):
		result, to_continue, err = await self.selected_authentication_context_server.authenticate_relay_server(token_data)
		if err is not None:
			return None, None, err
		if result is None:
			return None, False, None
		response = {}
		if include_negstate == True:
			if to_continue == True:
				response['negState'] = NegState('accept-incomplete')
			else:
				response['negState'] = NegState('accept-completed')
		
		if result is not None and len(result) > 0:
			response['responseToken'] = result
		return response, to_continue, None
		
	async def process_ctx_authenticate(self, token_data, *args, **kwargs):
		result, to_continue, err = await self.selected_authentication_context.authenticate(token_data, *args, **kwargs)
		if err is not None:
			return None, None, err
		if not result:
			return None, False, None
		response = {}		
		response['responseToken'] = result
		return response, to_continue, None
	
	def get_copy(self):
		return self
	
	def list_original_conexts(self):
		"""
		Returns a list of authentication context names available to the SPNEGO authentication.
		"""
		return list(self.authentication_contexts.keys())
	
	def get_original_context(self, ctx_name):
		"""
		Returns a copy of the original (not used) authentication context sp[ecified by name.
		You may use this ctx to perform future authentication, as it has the user credentials
		"""
		return self.authentication_contexts[ctx_name]

	def get_extra_info(self):
		if hasattr(self.selected_authentication_context, 'get_extra_info'):
			return self.selected_authentication_context.get_extra_info()
		return None
	
	def get_session_key(self):
		return self.selected_authentication_context.get_session_key()

	def get_mechtypes_list(self):
		neghint = {'hintName':'not_defined_in_RFC4178@please_ignore'}
		tokinit = {
			'mechTypes': [MechType(mt) for mt in self.authentication_contexts],
			'negHints': NegHints(neghint),
		}

		negtoken = NegotiationToken({'negTokenInit':NegTokenInit2(tokinit)})
		#spnego = GSS_SPNEGO({'NegotiationToken':negtoken})
		return GSSAPI({'type': GSSType('1.3.6.1.5.5.2'), 'value':negtoken}).dump()
	
	async def authenticate_relay_server(self, token, flags = None, seq_number = 0, is_rpc = False):
		self.__server_latest_token = token
		try:
			if self.__client_side_crash is not None:
				return None, False, Exception('Client auth crashed!')
			negtoken = None
			if self.selected_authentication_context_server is None:
				gss = GSSAPI.load(token).native
				negtoken = gss['value']
				if len(negtoken['mechTypes']) == 1:
					self.selected_mechtype_server = negtoken['mechTypes'][0]
					if negtoken['mechTypes'][0] == 'NTLMSSP - Microsoft NTLM Security Support Provider':
						await self.notify_relay() # something is happening here!
						self.selected_authentication_context_server = self.authentication_contexts[negtoken['mechTypes'][0]]


				else:
					raise Exception('This path is not yet implemented')

			if self.selected_authentication_context_server is not None:
				if negtoken is None:
					if token[0] == 0x60:
						gss = GSSAPI.load(token).native
						negtoken = gss['value']
					else:
						neg_token_raw = NegotiationToken.load(token)
						negtoken = neg_token_raw.native
				if 'mechToken' in negtoken:
					authdata = negtoken['mechToken']
				else:
					authdata = negtoken['responseToken']
				if 'mechListMIC' in negtoken:
					self.mic_data = negtoken['mechListMIC']
				response, to_continue, err = await self.process_ctx_authenticate_server(authdata, flags = flags, seq_number = seq_number, is_rpc = is_rpc, include_negstate = True)
				if err is not None:
					return None, None, err

				if response is None:
					return None, False, None
				if self.iteration_ctr_server == 0:
					response['supportedMech'] = MechType(self.selected_mechtype_server)
					negtoken = NegotiationToken({'negTokenResp':NegTokenResp(response)})
				if self.iteration_ctr_server == 1:
					negtoken = NegotiationToken({
						'negTokenResp' : NegTokenResp(response),
					})

				
				self.iteration_ctr_server += 1
				return negtoken.dump(), to_continue, None
		except Exception as e:
			self.__server_side_crash = e
			return None, False, e
	
	
	async def authenticate(self, token, *args, **kwargs): #seq_number = 0, is_rpc = False):
		"""
		This function is called (multiple times) during negotiation phase of a protocol to determine hich auth mechanism to be used
		Token is a byte array that is an ASN1 NegotiationToken structure.
		"""
		try:
			if self.__server_side_crash is not None:
				return None, False, Exception('Server auth crashed!')
			if self.selected_mechtype is None:
				if token is None:
					#first call to auth, we need to create NegTokenInit2
					#we must list all available auth types, if only one is present then generate initial auth data with it
					
					selected_name = None
					mechtypes = []
					for mechname in self.authentication_contexts:
						selected_name = mechname #only used if there is one!
						mechtypes.append(MechType(mechname))
					
					response = {}
					response['mechTypes'] = MechTypes(mechtypes)
					
					if len(mechtypes) == 1:
						self.selected_authentication_context = self.authentication_contexts[selected_name]
						self.selected_mechtype = selected_name
						result, to_continue, err = await self.selected_authentication_context.authenticate(None, *args, **kwargs)
						if err is not None:
							return None, None, err
						response['mechToken'] = result
						#if is_rpc == False:
						#	response['mechToken'] = result
						#else:
						#	if not result:
						#		return None, False, None
						#	if str(response['mechTypes'][0]) == '1.2.840.48018.1.2.2':
						#		response['mechToken'] = KRB5Token(result).to_bytes()
						#	else:
						#		raise Exception('NTLM as RPC GSSAPI not implemented!')
					
					### First message and ONLY the first message goes out with additional wrapping
					
					negtoken = NegotiationToken({'negTokenInit':NegTokenInit2(response)})
					
					
					#spnego = GSS_SPNEGO({'NegotiationToken':negtoken})
					return GSSAPI({'type': GSSType('1.3.6.1.5.5.2'), 'value':negtoken}).dump(), True, None
					
					
				else:
					#we have already send the NegTokenInit2, but it contained multiple auth types,
					#at this point server is replying which auth type to use
					neg_token_raw = NegotiationToken.load(token)
					neg_token = neg_token_raw.native
					
					if not isinstance(neg_token_raw, NegTokenResp):
						raise Exception('Server send init???')
						
					self.selected_authentication_context = self.authentication_contexts[neg_token.mechTypes[0]]
					self.selected_mechtype = neg_token['supportedMech']
	
					response, to_continue, err = await self.process_ctx_authenticate(neg_token['responseToken'], *args, **kwargs)
					if err is not None:
						return None, None, err
					return NegTokenResp(response).dump(), to_continue, None
					
			else:
				neg_token_raw = NegotiationToken.load(token)
				neg_token = neg_token_raw.native
				if neg_token['negState'] == 'accept-completed':
					return None, False, None
				if neg_token['responseToken'] is None:
					# https://tools.ietf.org/html/rfc4178#section-5
					# mechlistmic exchange happening at the end of the authentication
					print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!111111111111')
					#print(neg_token_raw.native)
					#return self.__server_latest_token, True, None
					return None, True, None
				
				else:
					
					response, to_continue, err = await self.process_ctx_authenticate(neg_token['responseToken'], *args, **kwargs)
					if err is not None:
						return None, None, err
					if not response:
						return None, False, None

					#if self.selected_mechtype.startswith('NTLM'):
					#	response['mechListMIC'] = await self.sign(self.negtypes_store, 0, reset_cipher = True)
					#	self.internal_seq += 1
					res = NegotiationToken({'negTokenResp':NegTokenResp(response)}).dump()

					if self.__authtype is None:
						return self.__server_latest_token, True, None
					else:
						return res, to_continue, None
					#return res, to_continue, None
			
				##everything is netotiated, but authentication needs more setps
				#neg_token_raw = NegotiationToken.load(token)
				#neg_token = neg_token_raw.native
				#if neg_token['negState'] == 'accept-completed' and neg_token['responseToken'] is None:
				#	return None, True, None
				#response, to_continue, err = await self.process_ctx_authenticate(neg_token['responseToken'], flags = flags, seq_number = seq_number, is_rpc = is_rpc)
				#if err is not None:
				#	return None, None, err
				#if not response:
				#	return None, False, None
				#
				#if self.mic_data is not None:
				#	response['mechListMIC'] = self.mic_data
				#return NegotiationToken({'negTokenResp':NegTokenResp(response)}).dump(), to_continue, None
		except Exception as e:
			self.__client_side_crash = e
			return None, False, e

def spnegorelay_ntlm_factory(auth_relay_queue:asyncio.Queue, ntlm_handler_factory: Callable = None) -> SPNEGORelay:
	ntlm_ctx = ntlm_handler_factory()
	gssapi = SPNEGORelay(auth_relay_queue)
	gssapi.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', ntlm_ctx)
	ntlm_ctx.spnego_obj = gssapi
	return gssapi